*---------------------------------------------------------------------------------------
* Title      : asic27a sim for Puzzli 2 / Puzzli 2 Super on IGS Poly GameMaster Hardware
* Written by : iq_132
* Date       : 9/2/21
* Description: reimplementation of asic 27a simulation from MAME in 68K assembly
*---------------------------------------------------------------------------------------

    ORG    $120000
    
START:                  ; first instruction of program

*---------------------------------------------------------------------------------------
;   defines

region					equ		$5						 ; world!

_ramloc					equ		$818000
_level_structure        equ     _ramloc+$000             ; 160 bytes!
_sim_value		        equ	    _ramloc+$100             ; 2 bytes
_response		        equ	    _ramloc+$104	         ; 4 bytes
_hack_47_value          equ     _ramloc+$108             ; 4 bytes
_command_31_write_type  equ     _ramloc+$10c             ; 4 bytes
_p2_31_retcounter       equ     _ramloc+$110             ; 4 bytes
_hackcount2             equ     _ramloc+$114             ; 4 bytes
_hackcount              equ     _ramloc+$118             ; 4 bytes
_key                    equ     _ramloc+$11c             ; 4 bytes
_realrow                equ     _ramloc+$120             ; 4 bytes
_object_value_stored    equ     _ramloc+$124             ; 2 bytes
_tableoffs              equ     _ramloc+$128             ; 4 bytes

; this order is important for clearing these later
_tableoffs2             equ     _ramloc+$12c             ; 4 bytes
_entries_left           equ     _ramloc+$130             ; 4 bytes
_currentcolumn          equ     _ramloc+$134             ; 4 bytes
_currentrow             equ     _ramloc+$138             ; 4 bytes
_num_entries            equ     _ramloc+$13c             ; 4 bytes
_numbercolumns          equ     _ramloc+$140             ; 4 bytes
_depth                  equ     _ramloc+$144             ; 4 bytes
_m_row_bitmask          equ     _ramloc+$148             ; 4 bytes
_stage                  equ     _ramloc+$14c             ; 4 bytes
; end

*---------------------------------------------------------------------------------------
;   jump table

_region_word
	dc.w		  region

    bra.s         _protection_routine_D0
;   bra.s         _protection_routine_fixed  ; fall through

*---------------------------------------------------------------------------------------

;   puzzli2/s has a hard-coded 'reset' command
_protection_routine_fixed
    movem.l     D0-D4/A0, -(A7)
    move.w	    #$ffc7, D0
    bra.s       _protection_routine

;   all other commands are passed through D0 register
_protection_routine_D0
    movem.l     D0-D4/A0, -(A7)
_protection_routine

;   key calculation - step 1
;   is the key passed by the 68k 'ff'?
    move.w      D0, D1
    lsr.w       #$8, D1
    cmp.b       #$ff, D1
    bne.s         _skip_calc_key_ff
    move.l      #$ffffffff, _key
_skip_calc_key_ff
    move.w      _key, D1

	; debug feature, write to protection emulation
 ;   move.w      _sim_value, $500000
 ;   move.w      D0, $500002

 ;  apply key to command and remove upper 'key' byte
    eor.b       D1, D0
    andi.w      #$ff, D0

;   apply key to sim value
	move.w      _sim_value, D2
	eor.w       D1, D2
	move.w      D2, _sim_value

;   set default asic response
	move.l		#$740000, _response

	; debug feature, write command to unmapped memory
;	move.w		D0, $fffffe

;   check for known commands and process them
    cmpi.b       #$13, D0
    beq         _13_routine
	cmpi.b		#$31, D0
	beq		    _31_routine
	cmpi.b		#$38, D0
	beq 		_38_routine
	cmpi.b		#$41, D0
	beq		    _41_routine
	cmpi.b       #$47, D0
	beq         _47_routine	
	cmpi.b		#$52, D0
	beq 		_52_routine
	cmpi.b       #$54, D0
	beq         _54_routine
	cmpi.b		#$61, D0
	beq		    _61_routine
	cmpi.b		#$63, D0
	beq		    _63_routine
	cmpi.b		#$67, D0
	beq		    _67_routine

;   complete asic response
_end_protection_routine
;   update key and expand it to 32 bits
    moveq.l     #$0, D0
    moveq.l     #$0, D1
    move.b      _key, D0
    addq.w       #$1, D0
    andi.w      #$ff, D0
    cmpi.b      #$ff, D0
    bne.s       _skip_reset_key
    moveq.l     #$1, D0
_skip_reset_key
    move.b      D0, D1
    lsl.w       #$8, D0
    move.b      D1, D0
    move.w      D0, D1
    swap        D0
    move.w      D1, D0
    move.l      D0, _key

	; debug feature, write result to unmapped memory for logging/comparing
;    move.l      _response, D2
;    move.w      D2, $ffe010
;    swap        D2
;    move.w      D2, $ffe018

;   apply key to response
    move.l      _response, D1
    eor.l       D0, D1

	; debug feature, read result from protection emulation
;   move.w      $500002, D1
;	swap		D1 
;   move.w      $500000, D1

    move.l      D1, _response

    movem.l     (A7)+, D0-D4/A0
	rts

*---------------------------------------------------------------------------------------
; commands

;   retreive data from level structure
_13_routine
    tst.w		_hackcount	;cmp.l       #$0, _hackcount
    bne.s       _13_not_0
	moveq.l		#$0, D0
    move.b      _depth, D0
    lsr.b       #$4, D0
    addq.w      #$1, D0
    or.l        #$2d0000, D0
    move.l      D0, _response
    addi.w       #$1, _hackcount
	bra.s         _end_protection_routine

_13_not_0
    move.l      _numbercolumns, D0
    mulu.w      #$a, D0
    addq.l      #$1, D0
    cmp.w       _hackcount, D0
    bgt.s       _get_leveldata

    move.w      #$1, _hackcount
    move.l      #$740054, _response
	bra         _end_protection_routine

_get_leveldata
	moveq.l		#$0, D0
    move.w      _hackcount, D0
    subq.w      #$1, D0
    add.w       D0, D0
    lea.l       _level_structure, A0
    move.w      (A0,D0.w), D0
    or.l        #$2d0000, D0
    move.l      D0, _response
    add.w       #$1, _hackcount
	bra         _end_protection_routine


;   build level structure
_31_routine
    cmpi.l      #$2, _command_31_write_type
    bne.s       _not_type2
    tst.b		_hackcount2	;cmp.l       #$0, _hackcount2
    bne.s       _not_hackcount0
    bsr         _take_leveldata ; ignore response
    moveq.l     #$0, D0
    move.w      _sim_value, D0
    andi.w      #$ff, D0
    move.b		#$1, _hackcount2	;addq.l      #$1, _hackcount2
    move.l      #$d20000, _response
	bra         _end_protection_routine
_not_hackcount0
    bsr         _take_leveldata ; response important
    cmp.b       #$0, D0
    bne.s       _end_is_not_0
    move.b		#$1, _hackcount2	;addq.l      #$1, _hackcount2
    move.l      #$d20000, _response
	bra         _end_protection_routine
_end_is_not_0
    clr.b		_hackcount2 ;move.l      #$0, _hackcount2
    move.l      _numbercolumns, D0
    andi.l      #$ffff, D0
    ori.l       #$630000, D0
    move.l      D0, _response
	bra         _end_protection_routine
_not_type2
    move.l      _p2_31_retcounter, D0
    addq.l       #$1, _p2_31_retcounter
    or.l        #$d20000, D0
    move.l      D0, _response
	bra         _end_protection_routine


;   'reset' command
_38_routine
    move.l      #$100, _key
	move.w		#$78, D0
	swap 		D0
	move.w		_region_word, D0
	lsl.w		#$8, D0
    move.l      D0, _response
	bra         _end_protection_routine


;   command 31 write type select
_41_routine
    move.l      #$740061, _response
    clr.l		_command_31_write_type	;move.l      #$0, _command_31_write_type
    bra         _end_protection_routine


;   command 52 data write
_47_routine
    move.l      #$740047, _response
	move.w		_sim_value, _hack_47_value
;    moveq.l     #$0, D0
;    move.w      _sim_value, D0
;    move.l      D0, _hack_47_value
    bra         _end_protection_routine


;   mangle data written to command 47
_52_routine
    move.l      #$740000, D0
    move.w      _hack_47_value, D0
    lsr.w       #$8, D0
    andi.b      #$f, D0
    mulu.w      #$19, D0
    cmp.w       #$0, _sim_value
    beq.s       _skip_52_adder
	moveq.l		#$0, D1
    move.w      _hack_47_value, D1
    andi.w      #$f, D1
    mulu.w      #$5, D1
    add.w       D1, D0
_skip_52_adder
    move.w      _sim_value, D1
    andi.w      #$f, D1
    add.w       D1, D0
    move.l      D0, _response
    bra         _end_protection_routine


;   reset level structure routine
_54_routine
    move.l      #$2, _command_31_write_type
    move.b      #$ff, _stage
    move.b      #$0, _hackcount2
    move.w      #$0, _hackcount
	moveq.l		#$0, D0
    move.b      #$27, D0    ; 160/4
    moveq.l      #$0, D1
    lea.l       _level_structure, A0
_clear_table_loop
    move.l      D1, (A0)+
    dbra        D0, _clear_table_loop
    move.l      #$360000, _response
    bra         _end_protection_routine


;   set command 31 write type and return counter
_61_routine
    move.l      #$360000, _response
    move.l      #$1, _command_31_write_type
    move.l      #$c, _p2_31_retcounter
    bra         _end_protection_routine


;   get z80 data offsets (first 8)
_63_routine
    move.w      _sim_value, D0
    andi.w      #$7, D0
    lsl.w       #$2, D0
    lea.l       _63_z80_table, A0
    cmp.w       #$3230, $100232 ; puzzli2s
    bne.s       _is_not_s_63
    	adda.l      #$20, A0
_is_not_s_63
    move.l      (A0,D0.w), _response
    bra         _end_protection_routine


;   get z80 data offsets (second 8)
_67_routine
    move.w      _sim_value, D0
    andi.w      #$7, D0
    lsl.w       #$2, D0
    lea.l       _67_z80_table, A0
    cmp.w       #$3230, $100232 ; puzzli2s
    bne.s       _is_not_s_67
    adda.l      #$20, A0
_is_not_s_67
    move.l      (A0,D0.w), _response
    bra         _end_protection_routine


*---------------------------------------------------------------------------------------
; support routines (command 31)


_get_realrow
    moveq.l     #$0, D0
    moveq.l     #$0, D1     ; loops
    moveq.l     #$0, D2     ; counter
    move.l      _currentrow, D3    ; bit_wanted
    add.l       #$1, D3
    move.l      _m_row_bitmask, D0

_realrow_loop1
    move.l      D0, D4
    lsr.l       #$1, D0
    andi.l      #$1, D4
    cmp.l       #$0, D4
    beq.s       _realrow_no_add
    
    addq.l      #$1, D2
    cmp.l       D3, D2
    bne.s       _realrow_no_add
    move.l      D1, _realrow
    rts
    
_realrow_no_add
    addq.l      #$1, D1
    cmpi.l      #$10, D1
    bne.s       _realrow_loop1
    move.l      #$ffffffff, _realrow
    rts

_take_leveldata
    cmpi.b      #$ff, _stage
    bne.s       _not_negative1_stage
    
    	move.w      _sim_value, D0
    	andi.l      #$ff, D0
    	move.b      D0, _tableoffs

		lea.l		_tableoffs2, A0
		clr.l		(A0)+ ;    	move.l      D0, _tableoffs2
    	clr.l		(A0)+ ;    	move.l      D0, _entries_left
    	clr.l		(A0)+ ;    	move.l      D0, _currentcolumn
    	clr.l		(A0)+ ;    	move.l      D0, _currentrow
    	clr.l		(A0)+ ;    	move.l      D0, _num_entries
    	clr.l		(A0)+ ;    	move.l      D0, _numbercolumns
    	clr.l		(A0)+ ;    	move.l      D0, _depth
    	clr.l		(A0)+ ;    	move.l      D0, _m_row_bitmask
    	clr.l		(A0)+ ;    	move.l      D0, _stage
    	rts
    
_not_negative1_stage

	lea.l		_tableoffs2, A0
    moveq.l     #$0, D0
    moveq.l     #$0, D1
    move.w      _sim_value, D0
    andi.w      #$ff, D0
    move.b      _tableoffs, D1
    add.b       (A0), D1
    addq.b      #$1, (A0)
    andi.b      #$f, (A0)
    lea.l       _puzzli2_level_decode, A0
    move.b      (A0,D1.w), D1
    eor.b       D1, D0

; stage 0

    tst.b		_stage	; cmpi.l      #$0, _stage
    bne.s       _not_stage0
    
    	; stage 0!!
    	move.b      #$1, _stage
    	move.b      D0, D1
    	andi.b      #$f0, D1
    	move.b      D1, _depth
    	andi.b      #$f, D0
    	addi.b      #$1, D0
    	move.l      D0, _numbercolumns    
    	moveq.l     #$0, D0
    	rts

_not_stage0

    cmpi.b      #$1, _stage
    bne.s       _not_stage1

    	move.b      #$2, _stage
    	move.l      D0, D1
    	lsr.l       #$4, D1
    	move.l      D1, _entries_left
    	move.l      D1, _num_entries
    	andi.b      #$f, D0
    	lsl.w       #$8, D0
    	move.l      D0, _m_row_bitmask
    	moveq.l     #$0, D0
    	rts

_not_stage1 ; stage 2

    cmpi.b      #$2, _stage
    bne.s       _not_stage2

    	move.b      #$3, _stage
    	or.l        D0, _m_row_bitmask
_do_entries_left_check          ; use in stage 3!!

    	tst.l	_entries_left	;cmp.l       #$0, _entries_left
    	bne.s       _skip_entry_check1
 
    		move.b      #$1, _stage
    		addq.l      #$1, _currentcolumn
    		clr.l	_currentrow	;move.l      #$0, _currentrow
   		 	clr.l	_m_row_bitmask	;move.l      #$0, _m_row_bitmask
    		move.l      _currentcolumn, D0
    		cmp.l       _numbercolumns, D0
    		bne.s       _skip_entry_check1
    		moveq.l     #$1, D0
    		rts

_skip_entry_check1
    moveq.l      #$0, D0
    rts
    
_not_stage2 ; stage 3

    cmpi.l      #$3, _stage
    bne         _not_stage3

    	lea.l       _object_value, A0
    	subq.l      #$1, D0
    	add.l       D0, D0
    	move.l      (A0,D0.w), D0
    	move.l      D0, _object_value_stored
 
   		bsr         _get_realrow    ;!!!
    	cmp.l       #$ffffffff, _realrow
    	beq.s       _skip_set_level_structure

    		lea.l       _level_structure, A0
    		move.l      _currentcolumn, D1
    		mulu.w      #$a, D1
    		add.l       _realrow, D1
    		add.l       D1, D1
    		move.w      _object_value_stored+$2, (A0,D1.w)

_skip_set_level_structure
    	addq.l      #$1, _currentrow
    	subq.l      #$1, _entries_left
    	bra         _do_entries_left_check

_not_stage3
    moveq.l      #$0, D0
    rts

*---------------------------------------------------------------------------------------
; data tables

; z80 offset table
_63_z80_table   dc.l    $1694a8, $16cfae, $16ebf2, $16faa8, $174416, $600000, $600000, $600000  ; puzzli2 - first 8
_63_z80_table_s dc.l    $19027a, $193D80, $1959c4, $19687a, $19b1e8, $600000, $600000, $600000  ; puzzli2s - first 8
_67_z80_table   dc.l    $166178, $166178, $166178, $166178, $166e72, $600000, $600000, $600000  ; puzzli2 - second 8
_67_z80_table_s dc.l    $18cf4a, $18cf4a, $18cf4a, $18cf4a, $18dc44, $600000, $600000, $600000  ; puzzli2s - second 8

; object value table
_object_value
    dc.w    $0100,$0101,$0102,$0103,$0104,$0105,$0106,$0107,$0108,$0109,$010a,$010b,$010c,$010d,$010e,$010f
    dc.w    $0110,$0120,$0121,$0122,$0123,$0124,$0125,$0126,$0127,$0128,$0129,$012a,$012b,$012c,$012d,$012e
    dc.w    $012f,$0130,$0140,$0141,$0142,$0143,$0144,$0145,$0146,$0147,$0148,$0149,$014a,$014b,$014c,$014d
    dc.w    $014e,$014f,$0150,$0180,$0181,$0182,$0183,$0184,$0185,$0186,$0187,$0188,$0189,$018a,$018b,$018c
    dc.w    $018d,$018e,$018f,$0190,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110
    dc.w    $0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110
    dc.w    $0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110
    dc.w    $0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110
    dc.w    $0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110
    dc.w    $0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110
    dc.w    $0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110
    dc.w    $0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110
    dc.w    $0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110
    dc.w    $0200,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110
    dc.w    $8000,$8020,$8040,$8060,$8080,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110
    dc.w    $0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110,$0110

; level decode - decrypt table
_puzzli2_level_decode
    dc.b    $32,$3e,$b2,$37,$31,$22,$d6,$0d,$35,$5c,$8d,$3c,$7a,$5f,$d7,$ac
    dc.b    $53,$ff,$eb,$44,$e8,$11,$69,$77,$d9,$34,$36,$45,$a6,$e9,$1c,$c6
    dc.b    $3b,$bd,$ad,$2e,$18,$df,$a1,$ab,$dd,$52,$57,$c2,$e5,$0a,$00,$6d
    dc.b    $67,$64,$15,$70,$b6,$39,$27,$78,$82,$d2,$71,$b9,$13,$f5,$93,$92
    dc.b    $fa,$e7,$5e,$b0,$f6,$af,$95,$8a,$7c,$73,$f9,$63,$86,$cb,$1a,$56
    dc.b    $f1,$3a,$ae,$61,$01,$29,$97,$23,$8e,$5d,$9a,$65,$74,$21,$20,$40
    dc.b    $d3,$05,$a2,$e1,$bc,$9e,$1e,$10,$14,$0c,$88,$9c,$ec,$38,$b5,$9d
    dc.b    $2d,$f7,$17,$0e,$84,$c7,$7d,$ce,$94,$16,$48,$a8,$81,$6e,$7b,$d8
    dc.b    $a7,$7f,$42,$e6,$a0,$2a,$ef,$ee,$24,$ba,$b8,$7e,$c9,$2b,$90,$cc
    dc.b    $5b,$d1,$f3,$e2,$6f,$ed,$9f,$f0,$4b,$54,$8c,$08,$f8,$51,$68,$c8
    dc.b    $03,$0b,$bb,$c1,$e3,$4d,$04,$c5,$8f,$09,$0f,$bf,$62,$49,$76,$59
    dc.b    $1d,$80,$de,$60,$07,$e0,$1b,$66,$a5,$be,$cd,$87,$dc,$c3,$6b,$4e
    dc.b    $d0,$fd,$d4,$3f,$98,$96,$2f,$4c,$b3,$ea,$2c,$75,$e4,$c0,$6c,$6a
    dc.b    $9b,$b7,$43,$8b,$41,$47,$02,$db,$99,$3d,$a3,$79,$50,$4f,$b4,$55
    dc.b    $5a,$25,$f4,$ca,$58,$30,$c4,$12,$a9,$46,$da,$91,$a4,$aa,$fc,$85
    dc.b    $fb,$89,$06,$cf,$fe,$33,$d5,$28,$1f,$19,$4a,$b1,$83,$f2,$72,$26

    END    START        ; last line of source

*~Font name~Courier New~
*~Font size~12~
*~Tab type~1~
*~Tab size~4~
